#!/usr/bin/env python                                                          
#coding: utf-8
# Adapted from findScalingV3.py
"""
Overview:
========
This tool finds a combination of linear scaling factors for two raw curves to fit a final curve.
The Q-values are expected to match

Required input arguments:
    *Q1*: Q-vector of the first dataset
    *I1*: intensity of the first dataset
    *E1*: relative intensity uncertainty of the first dataset
    *Q2*: Q-vector of the first dataset
    *I2*: intensity of the second dataset
    *E2*: relative intensity uncertainty of the second dataset
Optional input arguments: 
    *backgroundFit*: Boolean indicating whether or not to fit the background,
        Default: True
"""

__author__ = "Brian R. Pauw"
__contact__ = "brian.pauw@bam.de"
__license__ = "GPLv3+"
__date__ = "2020/04/01"
__status__ = "beta"

# correctionBase contains the __init__ function for these classes
from numpy import sqrt
import numpy as np
import pandas
from scipy.optimize import leastsq
import logging

class findLinearCombination(object):
    dataset1 = pandas.DataFrame()
    dataset2 = pandas.DataFrame()
    matchDataset = pandas.DataFrame()
    sc = None
    initialGuess = None

    def __init__(self, 
        dataset1 = None, 
        dataset2 = None, 
        matchDataset = None, 
        initialGuess = [1., 1.], 
        qmin = 0, qmax = np.inf):
        
        self.dataset1 = dataset1.copy()
        self.dataset2 = dataset2.copy()
        self.matchDataset = matchDataset.copy()
        self.initialGuess = initialGuess

        if self.dataset1 is None:
            logging.error("Dataset 1 cannot be None")
        if self.dataset2 is None:
            logging.error("Dataset 2 cannot be None")
        if self.matchDataset is None:
            logging.error("dataset to match cannot be None")
        # check Q
        assert (self.dataset2["Q"].shape == self.dataset1["Q"].shape), 'dataset Q vectors must have the same size'
        assert (self.dataset2["Q"] == self.dataset1["Q"]).any(), 'dataset Q vectors must match'
        
        # trimming to the q limits of interest
        for dataset in [dataset1, dataset2, matchDataset]:
            dataset = dataset.loc[(dataset.Q > qmin) & (dataset.Q < qmax)]

        self.sc = self.scale()

        self.dataset1Scaled = self.dataset1.copy()
        self.dataset2Scaled = self.dataset2.copy()
        self.dataset1Scaled.I *= self.sc[0]
        self.dataset1Scaled.IE *= self.sc[0]
        self.dataset2Scaled.I *= self.sc[1]
        self.dataset2Scaled.IE *= self.sc[1]
        # calculate curve deviation:
        self.dev = (self.matchDataset.I - 
                (self.dataset1Scaled.I + self.dataset2Scaled.I)) / self.matchDataset.I
        return 

    def scale(self):
        #match the curves
        sc = self.initialGuess

        sc, dummySuccess = leastsq(self.csqr, sc)

        conVal = self.csqrV1(sc)
        
        logging.info(
                f'findLinearCombination: scaling factors: {sc}, chi-squared: {conVal}')
        return sc

    def readScaledDatasets(self, sc):
        #returns all three datasets' intensity and uncertainties, scaled.
        If = self.matchDataset["I"]
        Ef = self.matchDataset["IE"]
        I1 = self.dataset1["I"] * sc[0]
        E1 = self.dataset1["IE"] * sc[0]
        I2 = self.dataset2["I"] * sc[1] 
        E2 = self.dataset2["IE"] * sc[1]
        return I1, E1, I2, E2, If, Ef

    def csqr(self, sc, useMask = True):
        #csqr to be used with scipy.optimize.leastsq
        I1, E1, I2, E2, If, Ef = self.readScaledDatasets(sc)

        return ((I1 + I2) - If) / (sqrt(E1**2 + E2**2 + Ef**2))

    def csqrV1(self, sc, useMask = True):
        #complete reduced chi-squared calculation
        I1, E1, I2, E2, If, Ef = self.readScaledDatasets(sc)

        return sum((((I1 + I2) - If) / (sqrt(E1**2 + E2**2 + Ef**2)) )**2) / np.size(If)

